/*
 * USBCAM abstraction library for USB webcam drivers
 * Version 0.1.1
 *
 * Copyright (C) 2007 Sam Revitch <samr7@cs.washington.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "usbcam_priv.h"

/*
 * This file contains the file_operations implementation
 *
 * TODO LIST:
 * - Add debug tracing to more ioctl paths
 */

/*
 * V4L file_operations callout implementations
 */

static int usbcam_v4l_open(struct inode *inode, struct file *filp)
{
	struct usbcam_dev *udp;
	struct usbcam_fh *ufp;
	int autopm_ref = 0;
	int work_ref = 0;
	int res = 0;

	/* The usbcam_dev is referenced by the videodev at this point */
	udp = container_of(video_devdata(filp), struct usbcam_dev, ud_vdev);

	ufp = (struct usbcam_fh *) kzalloc(sizeof(*ufp), GFP_KERNEL);
	if (!ufp)
		return -ENOMEM;

	ufp->ufh_dev = udp;
	videobuf_queue_init(&ufp->ufh_vbq,
			    &usbcam_videobuf_qops,
			    NULL,
			    NULL,
			    V4L2_BUF_TYPE_VIDEO_CAPTURE,
			    V4L2_FIELD_INTERLACED,
			    sizeof(struct usbcam_frame), ufp);

	mutex_lock(&udp->ud_open_lock);

	if (!udp->ud_user_refs) {
		res = usb_autopm_get_interface(udp->ud_intf);
		if (res)
			goto bail_nolock;
		autopm_ref = 1;
	}

	usbcam_lock(udp);

	if (udp->ud_disconnected) {
		res = -ENODEV;
		goto bail;
	}

	if (!udp->ud_user_refs) {
		if (!udp->ud_minidrv->um_ops->no_workref_on_open) {
			res = usbcam_work_ref(udp);
			if (res)
				goto bail;
			work_ref = 1;
		}

		if (usbcam_minidrv_op_present(udp, open)) {
			res = usbcam_minidrv_op(udp, open);
			if (res) {
				if (work_ref)
					usbcam_work_unref(udp);
				assert(!udp->ud_user_refs);	
				goto bail;
			}
		}

		/* Transfer the autopm reference */
		assert(autopm_ref);
		autopm_ref = 0;
	}

	udp->ud_user_refs++;
	filp->private_data = ufp;
	usbcam_get(udp);

bail:
	usbcam_unlock(udp);
bail_nolock:
	mutex_unlock(&udp->ud_open_lock);
	if (res)
		kfree(ufp);
	if (autopm_ref)
		usb_autopm_put_interface(udp->ud_intf);
	usbcam_work_maybe_stop(udp);
	return res;
}

static int usbcam_v4l_release(struct inode *inode, struct file *filp)
{
	struct usbcam_fh *ufp = (struct usbcam_fh *) filp->private_data;
	struct usbcam_dev *udp = ufp->ufh_dev;
	int autopm_ref = 0;

	videobuf_mmap_free(&ufp->ufh_vbq);

	mutex_lock(&udp->ud_open_lock);
	usbcam_lock(udp);

	assert(udp->ud_user_refs);
	if (udp->ud_excl_owner == filp)
		udp->ud_excl_owner = NULL;
	kfree(ufp);
	filp->private_data = NULL;

	if (!--udp->ud_user_refs) {
		usbcam_capture_stop(udp);

		if (usbcam_minidrv_op_present(udp, close))
			usbcam_minidrv_op(udp, close);
		if (!udp->ud_minidrv->um_ops->no_workref_on_open)
			usbcam_work_unref(udp);
		autopm_ref = 1;
	}

	usbcam_unlock(udp);
	mutex_unlock(&udp->ud_open_lock);
	if (autopm_ref)
		usb_autopm_put_interface(udp->ud_intf);
	usbcam_work_maybe_stop(udp);
	usbcam_put(udp);
	return 0;
}

static ssize_t usbcam_v4l_read(struct file *filp, char __user *data,
			       size_t count, loff_t *ppos)
{
	struct usbcam_fh *ufp = (struct usbcam_fh *) filp->private_data;
	ssize_t res;

	res = videobuf_read_one(&ufp->ufh_vbq, data, count, ppos,
				(filp->f_flags & O_NONBLOCK) ? 1 : 0);
	usbcam_work_maybe_stop(ufp->ufh_dev);
	return res;
}

static unsigned int usbcam_v4l_poll(struct file *filp,
				    struct poll_table_struct *wait)
{
	struct usbcam_fh *ufp = (struct usbcam_fh *) filp->private_data;

	return videobuf_poll_stream(filp, &ufp->ufh_vbq, wait);
}

static int usbcam_v4l_mmap(struct file *filp, struct vm_area_struct * vma)
{
	struct usbcam_fh *ufp = (struct usbcam_fh *) filp->private_data;

	return videobuf_mmap_mapper(&ufp->ufh_vbq, vma);
}

static void usbcam_dbg_v4l2_buffer_res(struct usbcam_dev *udp, int res,
				       void *arg, const char *prefix)
{
	struct v4l2_buffer *b __attribute__((unused)) =
		(struct v4l2_buffer *) arg;

	if (res) {
		usbcam_dbg(udp, IOCTL_BUF, "%s res:%d", prefix, res);
		return;
	}

	usbcam_dbg(udp, IOCTL_BUF, "%s out: index=%d type=%d bytesused=%d "
		   "flags=0x%x field=%d memory=%d m=0x%lx length=%d",
		   prefix, b->index, b->type, b->bytesused,
		   b->flags, b->field, b->memory, b->m.userptr, b->length);
}

static void usbcam_dbg_v4l2_pix_format(struct usbcam_dev *udp,
				       struct v4l2_pix_format *f,
				       const char *prefix)
{
	__u32 pixfmt = f->pixelformat;
	if (!pixfmt)
		pixfmt = 0x3f3f3f3f;
	usbcam_dbg(udp, IOCTL_FMT, "%s wid=%d hgt=%d fmt=%.4s field=%d "
		   "bpl=%d size=%d cs=%d", prefix,
		   f->width, f->height, (char *) &pixfmt, f->field,
		   f->bytesperline, f->sizeimage, f->colorspace);
}

static void usbcam_dbg_v4l2_pix_format_res(struct usbcam_dev *udp, int res,
					   struct v4l2_pix_format *f,
					   const char *prefix)
{
	if (res) {
		usbcam_dbg(udp, IOCTL_FMT, "%s %d", prefix, res);
		return;
	}
	usbcam_dbg_v4l2_pix_format(udp, f, prefix);
}

static int usbcam_v4l_int_ioctl(struct inode *inodep, struct file *filp,
				unsigned int cmd, void *arg)
{
	struct usbcam_fh *ufp = (struct usbcam_fh *) filp->private_data;
	struct usbcam_dev *udp = ufp->ufh_dev;
	int droplock = 0, res;

#ifdef CONFIG_VIDEO_V4L1_COMPAT
	/* V4L1 API -- Handle this before v4l_compat_translate_ioctl(). */
	if (cmd == VIDIOCGMBUF) {
		struct v4l2_requestbuffers req;
		struct video_mbuf *p = (struct video_mbuf *) arg;
		unsigned int i;

		/*
		 * APPBUG: motion keeps the first mmap, yet requests
		 * larger capture sizes.
		 */
		usbcam_lock(udp);
		ufp->ufh_flags |= USBCAM_FH_USE_FIXED_FB;
		usbcam_unlock(udp);

		req.type = ufp->ufh_vbq.type;
		req.count = 2;
		req.memory = V4L2_MEMORY_MMAP;
		res = videobuf_reqbufs(&ufp->ufh_vbq, &req);
		if (res) {
			usbcam_dbg(udp, IOCTL_BUF,
				   "VIDIOCGMBUF reqbufs failed: %d", res);
			return res;
		}

		p->frames = req.count;
		p->size = 0;
		for (i = 0; i < p->frames; i++) {
			p->offsets[i] = ufp->ufh_vbq.bufs[i]->boff;
			p->size += ufp->ufh_vbq.bufs[i]->bsize;
		}

		usbcam_dbg(udp, IOCTL_BUF, "VIDIOCGMBUF frames=%d size=%d",
			   p->frames, p->size);
		return 0;
	}
#endif

	/* If the command is any other V4L1 API, pass it to the translator */
	if (_IOC_TYPE(cmd) == 'v')
		return v4l_compat_translate_ioctl(inodep, filp, cmd, arg,
						  usbcam_v4l_int_ioctl);

	/*
	 * Pass 1: filter out ioctls that we don't want to give
	 * the minidriver a chance to handle.
	 * These tend to be related to frame management, which is
	 * none of the minidriver's business, and we do not allow the
	 * minidriver to intercept.
	 */

	switch (cmd) {
	case VIDIOC_REQBUFS: {
		struct v4l2_requestbuffers *r =
			(struct v4l2_requestbuffers *) arg;
		/* APPBUG: disable USE_FIXED_FB if we enter this path */
		usbcam_lock(udp);
		ufp->ufh_flags &= ~(USBCAM_FH_USE_FIXED_FB);
		usbcam_unlock(udp);

		usbcam_dbg(udp, IOCTL_BUF,
			   "VIDIOC_REQBUFS count=%d type=%d memory=%d",
			   r->count, r->type, r->memory);
		res = videobuf_reqbufs(&ufp->ufh_vbq, r);
		usbcam_dbg(udp, IOCTL_BUF, "REQBUFS result=%d", res);
		usbcam_work_maybe_stop(udp);
		return res;
	}
	case VIDIOC_QUERYBUF: {
		struct v4l2_buffer *b = (struct v4l2_buffer *) arg;
		usbcam_dbg(udp, IOCTL_BUF,
			   "VIDIOC_QUERYBUF in: index=%d type=%d",
			   b->index, b->type);
		res = videobuf_querybuf(&ufp->ufh_vbq, b);
		usbcam_dbg_v4l2_buffer_res(udp, res, b, "VIDIOC_QUERYBUF");
		usbcam_work_maybe_stop(udp);
		return res;
	}
	case VIDIOC_QBUF: {
		struct v4l2_buffer *b = (struct v4l2_buffer *) arg;
		/*
		 * APPBUG: ptlib / Ekiga has an issue with zeroing the
		 * flags field before calling QBUF.
		 *
		 * Minidriver support for fast input switching is
		 * unavailable for the time being.
		 */
		b->flags = 0;

		usbcam_dbg(udp, IOCTL_BUF, "VIDIOC_QBUF in: index=%d type=%d",
			   b->index, b->type);
		res = videobuf_qbuf(&ufp->ufh_vbq, b);
		usbcam_dbg_v4l2_buffer_res(udp, res, b, "VIDIOC_QBUF");
		usbcam_work_maybe_stop(udp);
		return res;
	}
	case VIDIOC_DQBUF: {
		struct v4l2_buffer *b = (struct v4l2_buffer *) arg;
		res = videobuf_dqbuf(&ufp->ufh_vbq, b,
				     (filp->f_flags & O_NONBLOCK) ? 1 : 0);
		usbcam_dbg_v4l2_buffer_res(udp, res, b, "VIDIOC_DQBUF");
		usbcam_work_maybe_stop(udp);
		return res;
	}
	case VIDIOC_STREAMON: {
		enum v4l2_buf_type f = *(int *) arg;

		if (f != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			usbcam_dbg(udp, IOCTL_BUF,
				   "VIDIOC_STREAMON: invalid buf type %d", f);
			return -EINVAL;
		}
		if (!udp->ud_excl_owner) {
			usbcam_lock(udp);
			if (!udp->ud_excl_owner)
				udp->ud_excl_owner = filp;
			usbcam_unlock(udp);
		}
		if (udp->ud_excl_owner != filp) {
			usbcam_dbg(udp, IOCTL_BUF,
				   "VIDIOC_STREAMON: not exclusive owner");
			return -EBUSY;
		}
		res = videobuf_streamon(&ufp->ufh_vbq);
		usbcam_dbg(udp, IOCTL_BUF, "VIDIOC_STREAMON: res:%d", res);
		usbcam_work_maybe_stop(udp);
		return res;
	}
	case VIDIOC_STREAMOFF: {
		enum v4l2_buf_type f = *(int *) arg;

		if (f != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			usbcam_dbg(udp, IOCTL_BUF,
				   "VIDIOC_STREAMOFF: invalid buf type %d", f);
			return -EINVAL;
		}
		res = videobuf_streamoff(&ufp->ufh_vbq);
		usbcam_dbg(udp, IOCTL_BUF, "VIDIOC_STREAMOFF: res:%d", res);
		usbcam_work_maybe_stop(udp);
		return res;
	}
	}

	/* Pass the ioctl to the minidriver, see if it takes responsibility */

	if (usbcam_minidrv_op_present(udp, ioctl)) {
		if (!udp->ud_minidrv->um_ops->unlocked_ioctl) {
			usbcam_lock(udp);
			droplock = 1;
		}
		res = usbcam_minidrv_op(udp, ioctl, cmd, arg);
		if (droplock)
			usbcam_unlock(udp);

		usbcam_work_maybe_stop(udp);

		if (res != -ENOIOCTLCMD)
			return res;
	}

	/*
	 * Pass 2: If the minidriver doesn't handle the ioctl, do something
	 * default.  The minidriver can override all of this stuff by
	 * intercepting.
	 */

	switch (cmd) {

	/* DEFAULT CAPABILITIES / DEVICE NAME / DRIVER NAME / BUS INFO */
	case VIDIOC_QUERYCAP: {
		struct v4l2_capability *cap = (struct v4l2_capability *) arg;

		usbcam_lock(udp);
		strlcpy(cap->driver,
			usbcam_drvname(udp->ud_minidrv),
			sizeof(cap->driver));
		strlcpy(cap->card, udp->ud_vdev.name, sizeof(cap->card));
		snprintf(cap->bus_info, sizeof(cap->bus_info),
			 "usb:%s", udp->ud_dev->dev.bus_id);
		cap->version = udp->ud_minidrv->um_version;
		cap->capabilities = (V4L2_CAP_VIDEO_CAPTURE |
				     V4L2_CAP_READWRITE |
				     V4L2_CAP_STREAMING);
		usbcam_unlock(udp);
		return 0;
	}

	/* DEFAULT FORMAT HANDLING - USE MINDIRIVER CALLOUTS */
	case VIDIOC_ENUM_FMT: {
		struct v4l2_fmtdesc *f = (struct v4l2_fmtdesc *) arg;
		struct usbcam_pix_fmt *pf;

		usbcam_lock(udp);

		res = -EINVAL;
		if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE)
			goto enum_fmt_done;
		if (f->index >= udp->ud_fmt_array_len)
			goto enum_fmt_done;

		res = 0;
		pf = (struct usbcam_pix_fmt *)
			&(((u8 *) udp->ud_fmt_array)
			  [f->index * udp->ud_fmt_array_elem_size]);
		f->flags = pf->flags;
		f->pixelformat = pf->pixelformat;
		strlcpy(f->description,
			pf->description,
			sizeof(f->description));

	enum_fmt_done:
		usbcam_unlock(udp);
		return res;
	}

	case VIDIOC_G_FMT: {
		struct v4l2_format *f = (struct v4l2_format *) arg;

		if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_G_FMT: invalid buf type %d",
				   f->type);
			return -EINVAL;
		}

		usbcam_lock(udp);
		f->fmt.pix = udp->ud_format;
		usbcam_unlock(udp);
		usbcam_dbg_v4l2_pix_format(udp, &f->fmt.pix,
					   "VIDIOC_G_FMT: res:");
		return 0;
	}

	case VIDIOC_S_FMT: {
		struct v4l2_format *f = (struct v4l2_format *) arg;

		usbcam_lock(udp);

		usbcam_dbg_v4l2_pix_format(udp, &f->fmt.pix,
					   "VIDIOC_S_FMT: param:");

		if (udp->ud_disconnected) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: device disconnected");
			res = -EIO;
			goto s_fmt_done;
		}
		if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: invalid buf type %d",
				   f->type);
			res = -EINVAL;
			goto s_fmt_done;
		}
		if (!udp->ud_excl_owner)
			udp->ud_excl_owner = filp;
		if (!memcmp(&f->fmt.pix, &udp->ud_format,
			    sizeof(f->fmt.pix))) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: nothing to do");
			res = 0;
			goto s_fmt_done;
		}
		if (!usbcam_minidrv_op_present(udp, set_format)) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: no minidriver op");
			res = -EINVAL;
			goto s_fmt_done;
		}
		if (udp->ud_excl_owner != filp) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: not exclusive owner");
			res = -EBUSY;
			goto s_fmt_done;
		}
		usbcam_capture_stop_nondestructive(udp);
		if (udp->ud_capturing) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_S_FMT: capture in progress");
			res = -EBUSY;
			goto s_fmt_done;
		}
		res = usbcam_minidrv_op(udp, set_format, &f->fmt.pix);
		usbcam_dbg_v4l2_pix_format_res(udp, res, &f->fmt.pix,
					       "VIDIOC_S_FMT: res:");

	s_fmt_done:
		usbcam_unlock(udp);
		return res;
	}

	case VIDIOC_TRY_FMT: {
		struct v4l2_format *f = (struct v4l2_format *) arg;

		usbcam_lock(udp);

		usbcam_dbg_v4l2_pix_format(udp, &f->fmt.pix,
					   "VIDIOC_TRY_FMT: param:");

		if (udp->ud_disconnected) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_TRY_FMT: device disconnected");
			res = -EIO;
			goto try_fmt_done;
		}
		if (f->type != V4L2_BUF_TYPE_VIDEO_CAPTURE) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_TRY_FMT: invalid buf type %d",
				   f->type);
			res = -EINVAL;
			goto try_fmt_done;
		}
		if (!usbcam_minidrv_op_present(udp, try_format)) {
			usbcam_dbg(udp, IOCTL_FMT,
				   "VIDIOC_TRY_FMT: no minidriver op");
			res = -EINVAL;
			goto try_fmt_done;
		}

		res = usbcam_minidrv_op(udp, try_format, &f->fmt.pix);
		usbcam_dbg_v4l2_pix_format_res(udp, res, &f->fmt.pix,
					       "VIDIOC_TRY_FMT: res:");

	try_fmt_done:
		usbcam_unlock(udp);
		usbcam_work_maybe_stop(udp);
		return res;
	}

	/* DEFAULT CONTROL HANDLING - USE MINIDRIVER ARRAY / CALLOUTS */
	case VIDIOC_QUERYCTRL: {
		struct v4l2_queryctrl *a = (struct v4l2_queryctrl *) arg;
		const struct usbcam_ctrl *ctrlp, *resp;
		u32 targ;
		int higher_ids = 0, res = -EINVAL;

		usbcam_lock(udp);
		droplock = 1;
		targ = a->id;

		resp = NULL;

		if (targ & V4L2_CTRL_FLAG_NEXT_CTRL) {
			/*
			 * Find the control with the least ID greater than or
			 * equal to a->id
			 */
			targ &= ~V4L2_CTRL_FLAG_NEXT_CTRL;
			list_for_each_entry(ctrlp, &udp->ud_ctrl_list,
					    uc_links) {
				if (ctrlp->uc_v4l.id <= targ) {
					if (!resp || (ctrlp->uc_v4l.id <
						      resp->uc_v4l.id))
						resp = ctrlp;
				}
			}

		} else {
			/* Find an exact match */
			list_for_each_entry(ctrlp, &udp->ud_ctrl_list,
					    uc_links) {
				if (ctrlp->uc_v4l.id == targ) {
					resp = ctrlp;
					break;
				}
				else if ((ctrlp->uc_v4l.id >=
					  V4L2_CID_PRIVATE_BASE) &&
					 (ctrlp->uc_v4l.id <
					  (V4L2_CID_PRIVATE_BASE + 1024)) &&
					 (targ >= V4L2_CID_PRIVATE_BASE) &&
					 (targ < ctrlp->uc_v4l.id)) {
					higher_ids = 1;
				}
			}

			if (!resp && higher_ids) {
				/*
				 * Deal with the private control enumeration
				 * nonsense that the CTRL_FLAG_NEXT_CTRL thing
				 * fixes.
				 */
				memset(a, 0, sizeof(*a));
				a->id = targ;
				a->type = V4L2_CTRL_TYPE_INTEGER;
				strlcpy(a->name, "Disabled", sizeof(a->name));
				a->flags = V4L2_CTRL_FLAG_DISABLED;
				res = 0;
			}
		}

		if (resp) {
			*a = resp->uc_v4l;
			res = 0;

			/* Fill in required values for certain types */
			switch (a->type) {
			case V4L2_CTRL_TYPE_BOOLEAN:
				a->minimum = 0;
				a->maximum = 1;
				a->step = 1;
				break;
			case V4L2_CTRL_TYPE_MENU:
				a->step = 1;
				break;
			default:
				break;
			}

			/*
			 * If a query function was provided, call it to
			 * postprocess the response structure, e.g. to set
			 * flags.
			 */
			if (resp->query_fn) {
				if (udp->ud_minidrv->um_ops->unlocked_ctrl) {
					usbcam_unlock(udp);
					droplock = 0;
				}
				res = resp->query_fn(udp, resp, a);
			}
		}

		if (droplock)
			usbcam_unlock(udp);
		usbcam_work_maybe_stop(udp);
		return res;
	}

	case VIDIOC_G_CTRL: {
		struct v4l2_control *a = (struct v4l2_control *) arg;
		const struct usbcam_ctrl *resp;
		struct v4l2_ext_control ec;
		int res;

		usbcam_lock(udp);
		droplock = 1;

		if (udp->ud_disconnected) {
			usbcam_unlock(udp);
			return -EIO;
		}

		resp = usbcam_ctrl_find(udp, a->id);
		if (!resp ||
		    (resp->uc_v4l.type == V4L2_CTRL_TYPE_BUTTON) ||
		    !resp->get_fn)
			res = -EINVAL;
		else {
			if (udp->ud_minidrv->um_ops->unlocked_ctrl) {
				usbcam_unlock(udp);
				droplock = 0;
			}
			memset(&ec, 0, sizeof(ec));
			ec.id = a->id;
			ec.value = a->value;
			res = resp->get_fn(udp, resp, &ec);
			memset(a, 0, sizeof(*a));
			a->id = ec.id;
			a->value = ec.value;
		}

		if (droplock)
			usbcam_unlock(udp);
		usbcam_work_maybe_stop(udp);
		return res;
	}

	case VIDIOC_S_CTRL: {
		struct v4l2_control *a = (struct v4l2_control *) arg;
		const struct usbcam_ctrl *resp;
		struct v4l2_ext_control ec;
		int res;

		usbcam_lock(udp);
		droplock = 1;

		if (udp->ud_disconnected) {
			usbcam_unlock(udp);
			return -EIO;
		}

		resp = usbcam_ctrl_find(udp, a->id);
		if (!resp) {
			res = -EINVAL;
		} else if (!resp->set_fn) {
			/* Read-only control */
			res = -EBUSY;
		} else if ((resp->uc_v4l.type != V4L2_CTRL_TYPE_BUTTON) &&
			   ((a->value < resp->uc_v4l.minimum) ||
			    (a->value > resp->uc_v4l.maximum))) {
			res = -ERANGE;
		} else {
			if (udp->ud_minidrv->um_ops->unlocked_ctrl) {
				usbcam_unlock(udp);
				droplock = 0;
			}	
			memset(&ec, 0, sizeof(ec));
			ec.id = a->id;
			ec.value = a->value;
			res = resp->set_fn(udp, resp, &ec);
			memset(a, 0, sizeof(*a));
			a->id = ec.id;
			a->value = ec.value;
		}

		if (droplock)
			usbcam_unlock(udp);
		usbcam_work_maybe_stop(udp);
		return res;
	}

	case VIDIOC_QUERYMENU: {
		struct v4l2_querymenu *a = (struct v4l2_querymenu *) arg;
		const struct usbcam_ctrl *resp;
		int res;

		usbcam_lock(udp);

		resp = usbcam_ctrl_find(udp, a->id);

		if (!resp ||
		    (resp->uc_v4l.type != V4L2_CTRL_TYPE_MENU) ||
		    (a->index > resp->uc_v4l.maximum)) {
			res = -EINVAL;
			goto querymenu_done;
		}

		strlcpy(a->name,
			resp->uc_menu_names[a->index],
			sizeof(a->name));
		res = 0;

	querymenu_done:
		usbcam_unlock(udp);
		return res;
	}

	/* DEFAULT INPUT HANDLING -- There is one input called "Camera" */
	case VIDIOC_ENUMINPUT: {
		const struct v4l2_input dfl_input = {
			.name = "Camera",
			.type = V4L2_INPUT_TYPE_CAMERA,
		};
		struct v4l2_input *inp = (struct v4l2_input *) arg;

		if (inp->index > 0)
			return -EINVAL;
		*inp = dfl_input;
		return 0;
	}

	case VIDIOC_G_INPUT: {
		unsigned int *i = (unsigned int *) arg;
		*i = 0;
		return 0;
	}

	case VIDIOC_S_INPUT: {
		unsigned int *i = (unsigned int *) arg;
		if (*i != 0)
			return -EINVAL;
		return 0;
	}

	/* DEFAULT VIDEO STANDARD HANDLING */
	case VIDIOC_ENUMSTD: {
		struct v4l2_standard *s = (struct v4l2_standard *) arg;
		if (s->index > 0)
			return -EINVAL;
		v4l2_video_std_construct(s, V4L2_STD_NTSC_M, "NTSC-M");
		return 0;
	}

	case VIDIOC_G_STD: {
		v4l2_std_id *std = (v4l2_std_id *) arg;
		*std = V4L2_STD_NTSC_M;
		return 0;
	}

	case VIDIOC_S_STD:
		return 0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
	/* Workaround for v4l1_compat bug in older kernels */
	case VIDIOC_G_FBUF: {
		struct v4l2_framebuffer *f = (struct v4l2_framebuffer *) arg;
		memset(f, 0, sizeof(*f));
		return -ENOIOCTLCMD;
	}
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20) */
	}

	/* video_usercopy() will convert this to -EINVAL. */
	return -ENOIOCTLCMD;
}

static int usbcam_v4l_ioctl(struct inode *inodep, struct file *filp,
			    unsigned int cmd, unsigned long arg)
{
	return video_usercopy(inodep, filp, cmd, arg, usbcam_v4l_int_ioctl);
}
	

/*
 * The template file_operations structure
 *
 * Each usbcam_minidrv_t contains its own copy of this, which
 * is associated with the video4linux device created for that
 * minidriver.
 *
 * In general, copies will differ only in the .owner field, which
 * will refer to the minidriver module, not usbcam.
 */

struct file_operations usbcam_v4l_fops_template = {
	.owner		= THIS_MODULE,
	.open		= usbcam_v4l_open,
	.release	= usbcam_v4l_release,
	.read		= usbcam_v4l_read,
	.poll		= usbcam_v4l_poll,
	.mmap		= usbcam_v4l_mmap,
	.ioctl		= usbcam_v4l_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl	= v4l_compat_ioctl32,
#endif
	.llseek		= no_llseek,
};
